#include <linux/futex.h>

#include "samples.h"

typedef u_int32_t futex_t;
typedef u_int32_t cas_t;

chronos_mutex_t *r1;
pthread_mutex_t *p1, *p2;
futex_t *f1;
cas_t *c1;

unsigned long long time1, time2, time3;
unsigned long long vt_ltimes[1000], vt_utimes[1000], pt1_ltimes[1000], pt1_utimes[1000];
unsigned long long pt2_ltimes[1000], pt2_utimes[1000], f_ltimes[1000], f_utimes[1000];
unsigned long long c_ltimes[1000], c_utimes[1000];

void do_futex(futex_t *f, int op, int val) {
	syscall(SYS_futex, f, op, val, NULL, NULL, 0, 0);
}

void futex_wait(futex_t *f, int val) {
	do_futex(f, FUTEX_WAIT, val);
}

void futex_wake(futex_t *f) {
	do_futex(f, FUTEX_WAKE, 1);
}

static void futex_init(futex_t *f) {
	*f = 0;
}

static void futex_lock(futex_t *f) {
	int c;
	if((c = __sync_val_compare_and_swap(f, 0, 1)) != 0) {
		do {
			if(c == 2 || __sync_val_compare_and_swap(f, 1, 2) != 1) {
				futex_wait(f, 2); }
		} while ((c = __sync_val_compare_and_swap(f, 0, 2)) != 0);
	}
}

static void futex_unlock(futex_t *f) {
	if(__sync_val_compare_and_swap(f, 1, 0) == 2) {
		*f = 0;
		futex_wake(f);
	}
}

inline void cas_init(cas_t *c) {
	*c = 0;
}

static void cas_lock(futex_t *c) {
	while( __sync_val_compare_and_swap(c, 0, 1) != 0) { }
}

inline void cas_unlock(cas_t *c) {
	*c = 0;
}

void task() {
	int i;

	for(i = 0; i < 1000; i++) {
		time1 = RDTSC();
		pthread_mutex_lock(p1);
		time2 = RDTSC();
		pthread_mutex_unlock(p1);
		time3 = RDTSC();

		pt1_ltimes[i] = time2 - time1;
		pt1_utimes[i] = time3 - time2;
	}

	for(i = 0; i < 1000; i++) {
		time1 = RDTSC();
		pthread_mutex_lock(p2);
		time2 = RDTSC();
		pthread_mutex_unlock(p2);
		time3 = RDTSC();

		pt2_ltimes[i] = time2 - time1;
		pt2_utimes[i] = time3 - time2;
	}

	for(i = 0; i < 1000; i++) {
		time1 = RDTSC();
		chronos_mutex_lock(r1);
		time2 = RDTSC();
		chronos_mutex_unlock(r1);
		time3 = RDTSC();

		vt_ltimes[i] = time2 - time1;
		vt_utimes[i] = time3 - time2;
	}

	for(i = 0; i < 1000; i++) {
		time1 = RDTSC();
		futex_lock(f1);
		time2 = RDTSC();
		futex_unlock(f1);
		time3 = RDTSC();

		f_ltimes[i] = time2 - time1;
		f_utimes[i] = time3 - time2;
	}

	for(i = 0; i < 1000; i++) {
		time1 = RDTSC();
		cas_lock(c1);
		time2 = RDTSC();
		cas_unlock(c1);
		time3 = RDTSC();

		c_ltimes[i] = time2 - time1;
		c_utimes[i] = time3 - time2;
	}
}

int main()
{
	struct stat_data pt1_ltime, pt1_utime, pt2_ltime, pt2_utime, vt_ltime, vt_utime, f_ltime, f_utime, c_ltime, c_utime;
	pthread_mutexattr_t mattr;

	if(prep_test())
		return 1;

	pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_ERRORCHECK);

	r1 = new chronos_mutex_t();
	p1 = new pthread_mutex_t();
	p2 = new pthread_mutex_t();
	f1 = new futex_t();
	c1 = new cas_t();
	chronos_mutex_init(r1);
	pthread_mutex_init(p1, NULL);
	pthread_mutex_init(p1, &mattr);
	futex_init(f1);
	cas_init(c1);

	task();

	selection_sort(pt1_ltimes, 1000);
	selection_sort(pt1_utimes, 1000);
	selection_sort(pt2_ltimes, 1000);
	selection_sort(pt2_utimes, 1000);
	selection_sort(vt_ltimes, 1000);
	selection_sort(vt_utimes, 1000);
	selection_sort(f_ltimes, 1000);
	selection_sort(f_utimes, 1000);
	selection_sort(c_ltimes, 1000);
	selection_sort(c_utimes, 1000);

	pt1_ltime = stats(pt1_ltimes, 990);
	pt1_utime = stats(pt1_utimes, 990);
	pt2_ltime = stats(pt2_ltimes, 990);
	pt2_utime = stats(pt2_utimes, 990);
	vt_ltime = stats(vt_ltimes, 990);
	vt_utime = stats(vt_utimes, 990);
	f_ltime = stats(f_ltimes, 990);
	f_utime = stats(f_utimes, 990);
	c_ltime = stats(c_ltimes, 990);
	c_utime = stats(c_utimes, 990);

	cout << "Pthread Locking\t\t" << stat_print_long(pt1_ltime) << endl;
	cout << "Pthread Unlocking\t" << stat_print_long(pt1_utime) << endl;
	cout << "Pthread (EC) Locking\t" << stat_print_long(pt2_ltime) << endl;
	cout << "Pthread (EC) Unlocking\t" << stat_print_long(pt2_utime) << endl;
	cout << "ChronOS Locking\t\t" << stat_print_long(vt_ltime) << endl;
	cout << "ChronOS Unlocking\t" << stat_print_long(vt_utime) << endl;
	cout << "Futex Locking\t\t" << stat_print_long(f_ltime) << endl;
	cout << "Futex Unlocking\t\t" << stat_print_long(f_utime) << endl;
	cout << "CAS Locking\t\t" << stat_print_long(c_ltime) << endl;
	cout << "CAS Unlocking\t\t" << stat_print_long(c_utime) << endl;

	chronos_mutex_destroy(r1);
	delete r1;
	pthread_mutex_destroy(p1);
	pthread_mutex_destroy(p2);
	delete p1;
	delete p2;
	delete f1;
	delete c1;

	return 0;
}

