package edu.vt.realtime.testing;

import edu.vt.realtime.*;
import java.io.PrintStream;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;

public class ChronosTester {
	private TesterOptions opt;
	
	private ChronosAborts aborts;  // Data structure to interact with scheduler aborts

	private ArrayList<Task> taskSet;

	private ArrayList<ChronosMutex> locks;
	
	/* System slope */
	private double slope;

	/* Holds a pessimistic estimate for how long it takes to lock one of our mutexes */
	private long lock_time;

	/* Run statistics */
	private Timespec g_start_time;
	private Timespec g_end_time;

	/* System level performance statistics */
	private int sys_total_release; // The total number of tasks
	private int sys_met_release;   // The number of tasks that met deadlines
	private int sys_total_util;    // The total utility of all tasks
	private int sys_met_util;      // The total utility of all tasks that met deadlines
	private int sys_abort_count;   // The number of threads aborted

	private int cpu_usage;         // The current % of CPU usage for this run

	private double utilization = 0;    //kept for the sole purpose of making the GUI happy


	public ChronosTester() {
		/*Runtime runtime = Runtime.getRuntime();
		int procCount = runtime.availableProcessors();*/
		
		//Initialize the abort device
		this.aborts = ChronosScheduler.initAborts();

		//Initialize the list of locks
		this.locks = new ArrayList<ChronosMutex>();
	}

	//methods to handle printing in a consistent manner
	private void println(String str) {
		if (this.opt != null)
			this.opt.println(str);
	}
	private void print(String str) {
		if (this.opt != null)
			this.opt.print(str);
	}
	private void error(String str) {
		if (this.opt != null)
			this.opt.error(str);
	}
	
	private void clearCounters() {
		sys_total_release = 0;
		sys_met_release = 0;
		sys_total_util = 0;
		sys_met_util = 0;
		sys_abort_count = 0;
	}

	private void addCpu(int cpu) { this.opt.domain_mask |= (int)Math.pow(2,cpu); }

	private class LockTimeThread extends Thread{
		public void run() {
			ChronosMutex cm = new ChronosMutex();
			ChronosUtil.setLinuxScheduler(LinuxSchedulerType.SCHED_FIFO, Priorities.TASK_RUN.getNativeId());
			cm.lock();
			cm.unlock();
			cm.destroy();
		}
	}

	private void calcLockTime() {
		LockTimeThread thread;
		Timespec start, end, tmp;
		int prio;
		long time = 0, tmp_time = 0;

		start = new Timespec();
		end = new Timespec();
		prio = ChronosUtil.getPriority();
		
		ChronosUtil.setLinuxScheduler(LinuxSchedulerType.SCHED_FIFO, Priorities.MAIN.getNativeId());

		for(int i=0; i < 100; i++) {
			start.getTime();
			
			thread = new LockTimeThread();
			thread.start();
			try{thread.join();}catch(Exception e){}

			end.getTime();

			tmp = end.copy();
			tmp.subtract(start);
			tmp_time = tmp.inNanoseconds();
			if(tmp_time > time)
				time = tmp_time;
		}

		this.lock_time = time/1000;

		ChronosUtil.setLinuxScheduler(LinuxSchedulerType.SCHED_NORMAL);
		ChronosUtil.setPriority(prio);
	}

	/* Prepares to run the task set.
	 * Returns the hyper period of the task set */
	private boolean latestOptionsLoadValid = false;
	public int prepareForRun(TesterOptions opt) {
		this.opt = opt;

		this.opt.loadSlope();

		int toReturn = this.readTaskSetFile();

		latestOptionsLoadValid = this.opt.valid();

		return toReturn;
	}

	/* See if the file is ready to be run */
	public boolean prepared() { return latestOptionsLoadValid && fileReadSuccessfully; }

	/* Returns the hyper period for the task set, if successful, and -1 otherwise*/
	private boolean fileReadSuccessfully = false;
	public int readTaskSetFile() {
		fileReadSuccessfully = false; //set flag to false, so it will remain so if we encounter an error and return prematurely

		long lcm = 0;
		int tasks = 0;

		//set the TaskSet list to an empty one
		this.taskSet = new ArrayList<Task>();
		//set the domain mask to 0
		this.opt.domain_mask = 0;

		BufferedReader in;
		try{
			in = new BufferedReader(new FileReader(this.opt.taskSetPath));
		} catch(Exception e) {
			error("The task set file could not be found or opened. The supplied file name was "+this.opt.taskSetPath);
			return -1;
		}

		String str;
		try{
		while ((str = in.readLine()) != null) {

			if(str.startsWith("L") && str.length() > 2) {
				try{
					this.opt.lock_count = Integer.parseInt(str.substring(2));
				}catch(Exception e){
					error("Failed to parse number of locks in task set file. Is the number formatted correctly?");
					return -1;
				}

				//if we have locks, set the option
				if(this.opt.lock_count > 0)
					this.opt.locking = true;
				else
					this.opt.locking = false;

			} else if(str.startsWith("T") && str.length() > 2) {
				tasks++;
				String[] list = str.split("\\s+");

				//make sure the task set is valid
				if(list.length < 4) {
					error("Task "+tasks+" in the task set is invalid.");
					return -1;
				}

				Task t = new Task(this.opt);

				//Set which CPUs the task can run on
				if(list[1].equals("all")) {
					for(int i=0; i < this.opt.proc_count; i++) {
						t.addCpu(i);
						addCpu(i);
					}
				} else {
					String[] cpus = list[1].split(",");
					for(int i=0; i < cpus.length; i++) {
						t.addCpu(Integer.parseInt(cpus[i]));
						addCpu(Integer.parseInt(cpus[i]));
					}
				}

				//set the period and execution time of the task
				t.setPeriods(Long.parseLong(list[2]));
				t.setExecTime(Long.parseLong(list[3]));

				//set the utility, if given
				if(list.length >= 5)
					t.setUtility(Integer.parseInt(list[4]));
				else
					t.setUtility(0);

				//setup the locks, if applicable
				if(list.length >= 6 && this.opt.lock_count > 0) {
					String[] lockList = list[5].split(",");
					for(int i=0; i < lockList.length; i++)
						t.addLock(Integer.parseInt(lockList[i]));
				}

				//setup abort utility, if applicable
				if(list.length >= 7) {
					t.setAbortUtility(Integer.parseInt(list[6]));
					if(t.getAbortUtility() > t.getUtility())
						println("Warning: task "+tasks+" has an abort utility greater than its normal utility.");
					this.opt.use_hua = true;
				}

				//finally, add the task to to task list
				taskSet.add(t);
			}
			//else, a comment or an invalid line, ignore it
		}
		in.close();
		} catch(Exception e) {
			error("An IOException occurred while reading the file " + this.opt.taskSetPath);
			e.printStackTrace();
			try{in.close();}catch(Exception f){}
			return -1;
		}

		//if we didn't add any tasks, fail
		if(taskSet.size() <=0) {
			error("The file you supplied didn't have any tasks.");
			return -1;
		}

		utilization = 0;
		//calculate the total CPU utilization for the task set
		for(int i=0; i < taskSet.size(); i++)
			utilization += taskSet.get(i).utilization();

		//calculate the hyper period for the task set
		lcm = taskSet.get(0).getPeriod();

		for(int i=1; i < taskSet.size(); i++) {
			int p = (int)taskSet.get(i).getPeriod();
			int GCD = (int)ChronosUtil.gcd(lcm, p);
			lcm = lcm/GCD;
			lcm = lcm * p;
		}
		
		fileReadSuccessfully = true; //do this last if everything succeeded

		return (int)lcm/1000000;
	}

	/* Functions to initialize and cleanup locks */
	private void initLocks() {
		if(this.opt.locking) {
			for(int i=0; i < this.opt.lock_count; i++)
				this.locks.add(new ChronosMutex());
		}
	}
	private void cleanupLocks() {
		while(this.locks.size() > 0) {
			ChronosMutex lock = this.locks.remove(0);
			lock.destroy();
		}
	}

	/* returns true on success, false otherwise */
	public boolean run() {
		if (this.opt.use_hua)
			calcLockTime();

		for(cpu_usage = this.opt.cpu_usage;
				cpu_usage <= this.opt.cpu_usage_end;
				cpu_usage += this.opt.cpu_usage_it) {
			boolean runSuccessful = doRun();

			if(cpu_usage <= this.opt.cpu_usage_end)
				try{Thread.sleep(1000);}catch(Exception e){}

			//flush the output from the previous runs
			this.opt.flush();

			if (!runSuccessful)
				return false;
		}
		return true;
	}

	/* returns true if successful, false otherwise */
	private boolean doRun() {
		boolean skip_run = false;
		int old_prio = 0;

		clearCounters();

		old_prio = ChronosUtil.getPriority();

		if (!ChronosUtil.setLinuxScheduler(LinuxSchedulerType.SCHED_FIFO, Priorities.MAIN.getNativeId())) {
			skip_run = true;
			error("sched_setscheduler failed, are you root (or sudo)?");
		}

		if (!ChronosUtil.setAffinity((int)Math.pow(2,this.opt.main_cpu))){
			skip_run = true;
			error("sched_setaffinity failed, are you root (or sudo)?");
		}

		if (!ChronosScheduler.setScheduler(this.opt.scheduler, Priorities.TASK_RUN.getNativeId(), this.opt.domain_mask)) {
			skip_run = true;
			error("Selecting ChronOS scheduler ("+this.opt.scheduler.toString().substring(9)+") failed! Is the scheduler loaded, and are you root?");
		}

		if(skip_run) {
			ChronosUtil.setLinuxScheduler(LinuxSchedulerType.SCHED_NORMAL, old_prio);
			return false;
		}

		this.initLocks(); //initialize the locks

		if(this.opt.verbose) {
			println("Percent of exec time to use per task: "+(double)this.opt.cpu_usage/100);
			println("Critical section length: "+(double)this.opt.cs_len/100);
			println("CPU: "+this.opt.main_cpu);
		}

		g_start_time = new Timespec();
		g_end_time = new Timespec();

		for(int i = 0; i < this.taskSet.size(); i++) {
			Task tsk = this.taskSet.get(i);
			tsk.initTask(i, this.locks, this.aborts, lock_time);
			tsk.setStartTime(g_start_time);
			this.sys_total_release += tsk.getRunReleases();
			this.sys_total_util += tsk.getRunUtility();

			if(this.opt.verbose) {
				println("period: "+tsk.period());
				println("usage in usec: "+tsk.unlocked_usage()+" unlocked + "+tsk.locked_usage()+" locked");
			}
		}

		g_start_time.getTime(); //get the current realtime clock

		for(int i=0; i < taskSet.size(); i++)
			taskSet.get(i).start();

		g_end_time.setTime(g_start_time.getSeconds() + this.opt.run_time, g_start_time.getNanoseconds());

		//Sleep while the tasks run, then harvest their statistics
		boolean runFailed = false;
		for(int i=0; i < taskSet.size(); i++) {
			Task tsk = taskSet.get(i);
			try{
				tsk.join();
			} catch(Exception e) {
				error("FAILED TO JOIN TASK #"+i);
				error("Numbers for this run will therefore be incorrect");
				runFailed = true;
				continue;
			}

			this.sys_met_release += tsk.deadlines_met();
			this.sys_met_util += tsk.utility_accrued();
			this.sys_abort_count += tsk.jobs_aborted();
		}

		this.cleanupLocks(); //free the locks

		ChronosUtil.setLinuxScheduler(LinuxSchedulerType.SCHED_NORMAL, old_prio); //return to normal scheduler and priority

		if (runFailed)
			println("Note: the following statistics are probably invalid, because at least one of the tasks was not properly joined and its statistics incorporated.");

		//print results out in whichever way we are told to
		if(this.opt.verbose) {
			println("set start_time: "+g_start_time);
			println("set end_time: "+g_end_time);
			println("Total tasks: " + sys_total_release +
				       	", total deadlines met: "+sys_met_release +
				       	", total possible utility : "+sys_total_util +
				       	", total utility accrued: "+sys_met_util +
				       	", total tasks aborted: "+sys_abort_count);
		} else if (this.opt.excel) {
			println(this.opt.scheduler.toString().substring(9) + "," + cpu_usage + 
					",=" + sys_met_release + "/" + sys_total_release + ",=" +
					sys_met_util + "/" + sys_total_util);
		} else if (this.opt.gnuplot) {
			println((double)cpu_usage/100 + " " + dsr_per() + " " + aur_per());
		} else {
			println((double)cpu_usage/100 + "/" + this.opt.scheduler.toString().substring(9) + 
					": Tasks: " + sys_met_release + "/" + sys_total_release + 
					", Utility: " + sys_met_util + "/" + sys_total_util +
					", Aborted: " + sys_abort_count);
		}

		return !runFailed; //hopefully everything went swimmingly
	}

	private double dsr_per() { return (double)sys_met_release / (double)sys_total_release; }
	private double aur_per() { return (double)sys_met_util / (double)sys_total_util; }

	public double getUtilization() { return utilization; }
	public int getNumTasks() { return this.taskSet.size(); }
}
