package edu.vt.realtime.testing;

import gnu.getopt.*;
import edu.vt.realtime.*;

public class SchedTestApp {

	private String optstring = "c:e:f:i:l:r:s:abdghmnopvxz";

	public static void main(String[] args) {
		//lock all pages of memory in physical memory
		if(!ChronosUtil.memoryLock()) {
			System.out.println("Error: Unable to lock the memory space");
			System.out.println("Are you running as root?");
			return;
		}
		
		SchedTestApp sta = new SchedTestApp();
		if(args.length == 0)
			sta.startGUI();
		else
			sta.startCLI(args);
	}

	private void printUsage() {
		System.out.println();
		System.out.println("Console testing application for ChronOS");
		System.out.println("------------------------------------------------------ ");
		System.out.println("Usage:");
		System.out.println("  -s scheduler  The name of scheduler - RMA, EDF, DASA etc");
		System.out.println("  -a            Abort on idle");
		System.out.println("  -p            Enable priority inheritance");
		System.out.println("  -h            Enable HUA abort handlers");
		System.out.println("  -d            Enable deadlock prevention");
		System.out.println("  -f filename   The name of the file that has the task list");
		System.out.println("  -b            Enable the batch mode");
		System.out.println("  -c cpu-usage  The cpu-usage");
		System.out.println("  -e end-usage  The end cpu-usage (used in batch mode)");
		System.out.println("  -i interval   The interval for each iteration in batch");
		System.out.println("  -r run-time   The total time the test needs to run");
		System.out.println("                This should be the hyper-period (use -z to find)");
		System.out.println("  -l cs-length  Enable locking, and provide critical");
		System.out.println("                section length");
		System.out.println("  -n            Enable nested locking (as opposed to sequential)");
		System.out.println("  -v            Enable verbose mode");
		System.out.println("  -o            Enable output to a log file");
		System.out.println("  -x            Enable log in Excel format");
		System.out.println("  -g            Enable log in Gnuplot format");
		System.out.println("  -z            Don't run the test, just print the hyper-period");
		System.out.println();
		System.out.println("Examples:");
		System.out.println("    $sched_test_app -s RMA -f taskset/5t_nl -c 100 -r 20");
		System.out.println(" OR $sched_test_app -s EDF -f taskset/10t_nl -c 50 -l 10 -r 20 --a -p -h -d");
		System.out.println();
		System.out.println(" NOTE: -a -p -h -d can also be written as -aphd");
	}

	public void startGUI() {
		TestAppGui gui = new TestAppGui();
	}

	private String[] fixFilenameArguments(String[] args) {
		int j=0;
		boolean inFilename = false; //true if we are after the -f and before the next argument with a preceeding "-"
		boolean justBeganFilename = false; //true if we are in the first 
		for(int i=0; i < args.length; i++) {
			//if we are no longer in the filename, break us out of it
			if (inFilename && args[i].startsWith("-")) {
				inFilename = false;
				j++;
			}

			if(inFilename){
				//update the strings, removing backslashes, adding spaces
				if(justBeganFilename) {
					if(args[i].endsWith("\\"))
						args[j] = args[i].substring(0, args[i].length()-1);
				} else if(args[i].endsWith("\\"))
					args[j] += " "+args[i].substring(0, args[i].length()-1);
				else
					args[j] += " "+args[i];
			} else {
				//simply set the strings equal, since we aren't in the filename section
				args[j] = args[i];
				j++;
			}

			justBeganFilename = false;
			//if we are entering a filename, set us to that mode
			if(args[i].equals("-f")) {
				inFilename = true;
				justBeganFilename = true;
			}
		}

		//now, fix it up so we don't have extra elements things on the end of the array
		String[] argsnew = new String[j];
		for(int i = 0; i < j; i++)
			argsnew[i] = args[i];

		return argsnew;
	}

	public void startCLI(String[] args){
		//Fix up the arguments for the case where we have a filename with backslashes in it
		args = fixFilenameArguments(args);

		//Display an error message if they did it wrong
		if(args.length < 7) {
			System.out.println("\nError: You must provide the following parameters at a minimum: the scheduler, CPU usage, run time, and filename for the taskset.\n");
			System.out.println("If you wish to run with a GUI, run this command with 0 arguments.\n");
			printUsage();
			return;
		}

		TesterOptions opt = new TesterOptions();

		Getopt g = new Getopt("SchedTestApp.java", args, optstring);
		int c;
		String arg = "";
		while ((c = g.getopt()) != -1) {
			arg = g.getOptarg();
			switch(c) {
			       case 'b' :
					opt.batch_mode = true;
					break;

				case 's' :
					ChronosSchedulerType t = ChronosSchedulerType.fromString(arg);
					if(t==null) {
						System.out.println("Error: Scheduler (\""+arg+"\") is not a valid ChronOS scheduler.");
						System.out.println("Valid Schedulers:");
						for (ChronosSchedulerType s: ChronosSchedulerType.values())
							System.out.println("    "+s.toString().substring(9));
						return;
					}
					opt.scheduler = t;
					break;

				case 'f' :
					if(arg != null && !arg.equals(""))
						opt.taskSetPath = arg;
					break;

				case 'c' :
					opt.cpu_usage = Integer.parseInt(arg);
					break;

				case 'e' :
					opt.cpu_usage_end = Integer.parseInt(arg);
					break;

				case 'i' :
					opt.cpu_usage_it = Integer.parseInt(arg);
					break;

				case 'l' :
					opt.locking = true;
					opt.cs_len = Integer.parseInt(arg); //Critical section length
					break;

				case 'r' :
					opt.run_time = Integer.parseInt(arg);
					break;

				case 'a' :
					opt.abort_idle = true;
					break;

				case 'p' :
					opt.use_pi = true;
					break;

				case 'g' :
					opt.gnuplot = true;
					break;

				case 'h' :
					opt.use_hua = true;
					break;

				case 'd' :
					opt.deadlock_prevention = true;
					break;

				case 'm' :
					opt.main_cpu = Integer.parseInt(arg);
					break;

				case 'n' :
					opt.nested_locking = true;
					break;

				case 'v' :
					opt.verbose = true;
					break;

				case 'o' :
					opt.log_to_file = true;
					break;

				case 'x' :
					opt.excel = true;
					break;

				case 'z' :
					opt.no_run = true;
					break;

				case '?' :
					break; //getopt() already printed out an error
				default :
					printUsage();
			}
		}

		if (opt.taskSetPath.equals("")) {
			System.out.println("\nError: Must supply task set file.");
			printUsage();
		}

		ChronosTester tester = new ChronosTester(); //create the tester

		//prepare to make a run
		int lcm = tester.prepareForRun(opt);

		if(!tester.prepared()) {
			System.out.println("Error: Cannot proceed due to invalid configuration");
			return;
		}
		
		//if we're not actually supposed to run, display Hyper-period and quit
		if(opt.no_run) {
			System.out.println("Hyper-period is: " + lcm + " s");
			return;
		}

		tester.run(); //actually run the tests
	}
}
