package edu.vt.realtime.testing;

import java.io.PrintStream;
import java.io.BufferedReader;
import java.io.FileReader;

import edu.vt.realtime.ChronosSchedulerType;

public class TesterOptions {

	/* PrintStream to print output to */
	PrintStream out = System.out;
	PrintStream err = System.err;

	String taskSetPath = "";
	String slopeFileName = "/usr/local/chronos/slope_java.conf";

	/* CPU Affinity */
	int proc_count = 0;
	int main_cpu = 0;
	long domain_mask = 0;

	/* CPU Usage */
	int cpu_usage = 0;
	int cpu_usage_end = 0;
	int cpu_usage_it = 0;

	/* Other run options */
	int run_time = 0;
	int cs_len = 0; //length of critical section
	
	int lock_count = 0; //number of locks
	boolean locking = false; //if locking is enabled
	boolean nested_locking = false; //do nested locking, instead of flat locking

	boolean abort_idle = false;
	boolean use_hua = false;
	boolean use_pi = false;
	boolean deadlock_prevention = false;

	boolean verbose = false;         //verbose output
	boolean excel = false;           //log to Excel file
	boolean gnuplot = false;         //log in GNU Plot format
	boolean log_to_file = false;     //log to an actual file

	boolean batch_mode = false;

	boolean no_run = false; //don't actually run test, only calculate and print hyper-period

	/* System slope */
	double slope = 0;

	/* Scheduler options */
	ChronosSchedulerType scheduler = ChronosSchedulerType.SCHED_RT_FIFO;

	private String toPrint = "";
	private String toError = "";

	public TesterOptions() {
		Runtime runtime = Runtime.getRuntime();
		this.proc_count = runtime.availableProcessors();
	}

	/* Methods to handle printing from within methods which are not part of the GUI */
	public synchronized void print(String str) {
		toPrint += str;
	}
	public void println(String str) {
		print(str+"\n");
	}
	public synchronized void error(String str) {
		toError += str+"\n";
	}
	public synchronized void flush() {
		if(this.out != null)
			this.out.print(toPrint);
		if(this.err != null)
			this.err.print(toError);
		else if (this.out != null)
			this.out.print(toError);
		toPrint = "";
		toError = "";
	}

	/* Attempt to load the slope from the slope file */
	public boolean loadSlope() {
		slope = 0; //reset it so it'll trigger valid() if we fail, instead of reporting a possibly old value
                BufferedReader in;
                try{
                        in = new BufferedReader(new FileReader(slopeFileName));
                } catch(Exception e) {
			if (err != null)
				err.println("Error: The slope file could not be found or opened. The supplied file name was "+slopeFileName);
			else
				System.err.println("Error: The slope file could not be found or opened. The supplied file name was "+slopeFileName);
                        return false;
                }

                try{
			String str = in.readLine();
			if(str == null)
				return false;
			this.slope = Double.parseDouble(str);
		} catch (Exception e) {
			return false;
		}
		return true;
	}

	public boolean valid() {
		boolean verified = true;

		if (out == null) {
			System.err.println("Error: Output PrintStream not provided");
			return false;
		}
		if (err == null) {
			System.err.println("Warning: Error PrintStream not provided. Using output PrintStream for errors. Using output PrintStream for errors.");
			err = out;
		}

		//fix up variables for batch_mode if we're not actually using it
		if(!batch_mode) {
			cpu_usage_end = cpu_usage;
			cpu_usage_it = 1;
		}
		
		if (taskSetPath.equals("")) {
			err.println("Must supply a filename for the task set");
			verified = false;
		}

		if(slope==0) {
			err.println("Slope must not be zero.");
			verified = false;
		}

		if(lock_count <=0 && locking) {
			err.println("Task set does not have any locks, but locking is enabled.");
			verified = false;
		}
		
		if(excel && verbose) {
			err.println("Cannot select both Verbose Output and Excel options.");
			verified = false;
		}

		if(cpu_usage <= 0 || cpu_usage > 10000) {
			err.println("Invalid CPU usage.");
			verified = false;
		}

		if(batch_mode && cpu_usage_end > 10000) {
			err.println("Invalid End CPU usage.");
			verified = false;
		}

		if(run_time <= 0 || run_time > 3600) {
			err.println("Invalid execution time.");
			verified = false;
		}

		if((cs_len <= 0 && locking) || cs_len > 100) {
			err.println("Invalid critical section length.");
			verified = false;
		}
		
		if(batch_mode && cpu_usage_end < cpu_usage) {
			err.println("Batchmode selected, but end usage is less than start usage.");
			verified = false;
		}

		if(batch_mode && cpu_usage_it == 0) {
			err.println("CPU usage iteration cannot be 0.");
			verified = false;
		}

		return verified;
	}
}
