/***************************************************************************
 *   Copyright (C) 2009-2011 Virginia Tech Real-Time Systems Lab           *
 *                                                                         *
 *   Written by Aaron Lindsay                                              *
 *   aaron@aclindsay.com                                                   *
 *                                                                         *
 *   Based on code written by Matthew Dellinger                            *
 *   mdelling@vt.edu                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "slope.h"

/*
 * Execute the workload for a certain number of iterations, and return the
 * number of nanoseconds it took.
 */
static unsigned long measure_workload(struct workload *w,
				      unsigned long iterations)
{
	int prev_scheduler, prev_prio;
	struct sched_param param;
	unsigned long prev_mask, mask;
	struct timespec starttime, endtime;
	void *workload_data = 0;

	//initialize the per-task data section of the workload
	if (w->init_task)
		workload_data = w->init_task();

	//get the current affinity for this process
	sched_getaffinity(0, sizeof(prev_mask), (cpu_set_t *) & prev_mask);
	//pin process to CPU 0
	MASK_ZERO(mask);	//clear the CPU mask
	MASK_SET(mask, 0);	//add CPU 0 to the CPU mask
	sched_setaffinity(0, sizeof(mask), (cpu_set_t *) & mask);

	//get the current scheduler and priority
	prev_scheduler = sched_getscheduler(0);	//get the previous scheduler, so we can go back to that when we're done
	prev_prio = getpriority(PRIO_PROCESS, 0);
	//set us as the highest real-time priority
	param.sched_priority = 99;
	sched_setscheduler(0, SCHED_FIFO, &param);

	clock_gettime(CLOCK_THREAD_CPUTIME_ID, &starttime);
	w->do_work(workload_data, iterations);
	clock_gettime(CLOCK_THREAD_CPUTIME_ID, &endtime);

	//set affinity to what it was before we set it to CPU 0
	sched_setaffinity(0, sizeof(prev_mask), (cpu_set_t *) & prev_mask);

	//set priority and scheduler to what it was before
	param.sched_priority = prev_prio;
	sched_setscheduler(0, prev_scheduler, &param);

	//cleanup the per-task data section of the workload
	if (w->cleanup_task)
		w->cleanup_task(workload_data);

	//convert the differrence in times to nanoseconds and return it
	return timespec_subtract_ns(&starttime, &endtime);
}

/*
 * Find the worst-case slope. Slope is defined as the number of iterations per
 * microsecond. For example, a slope of 10 means that this particular workload
 * on this particular machine can always execute 10 iterations in less than one
 * microsecond.
 */
static void update_wcet_slope(const int workload, const int verbose)
{
	unsigned long run_nsec, iterations_per_second;
	double estimated_slope;
	int i;
	double wcet_slope;
	struct workload *w = get_workload_struct(workload);

	//initialize the workload global data
	if (w->init_global)
		w->init_global();

	//run workload once for a fairly large number of iterations, and use this
	//number of iterations to estimate a very loose slope number
	run_nsec = measure_workload(w, DEFAULT_SLOPE_ITERS);
	estimated_slope = (1000.0 * DEFAULT_SLOPE_ITERS) / run_nsec;	//one thousand is the conversion from nanoseconds to microseconds

	//refine this iteration once, and warm up the workload by running it for approximately a tenth of a second
	iterations_per_second = MILLION * estimated_slope;
	run_nsec = measure_workload(w, iterations_per_second);
	estimated_slope = (1000.0 * iterations_per_second) / run_nsec;	//one thousand is the conversion from nanoseconds to microseconds

	//run the workload a bunch of times, and take the worst of the resulting slopes
	//TODO make this really WCET by adding a preempting higher-priority task on the same core, which is also a cache-thrasher
	iterations_per_second = MILLION * estimated_slope;
	wcet_slope = -1.0;
	for (i = 0; i < SLOPE_GEN_SAMPLES; i++) {
		double slope;
		unsigned long iterations =
		    (SLOPE_GEN_MIN_TIME +
		     i * ((SLOPE_GEN_MAX_TIME - SLOPE_GEN_MIN_TIME) /
			  SLOPE_GEN_SAMPLES)) * estimated_slope;
		run_nsec = measure_workload(w, iterations);
		slope = (1000.0 * iterations) / run_nsec;
		if (slope < wcet_slope || wcet_slope < 0.0)
			wcet_slope = slope;
		if (verbose) {
			printf
			    ("\t\titers: %ld, us: %ld, slope: %f (worst: %f)\n",
			     iterations, run_nsec / 1000, slope, wcet_slope);
		}
	}

	//tear down the global workload data
	if (w->cleanup_global)
		w->cleanup_global();

	//decrease this final slope by 1% just to be on the safe side
	wcet_slope = wcet_slope * 0.99;

	if (put_workload_slope(workload, TIMING_WCET, wcet_slope))
		fatal_error("Unable to write wcet slope");
}

/*
 * Find the average-case slope. Slope is defined as the number of iterations per
 * microsecond. For example, a slope of 10 means that this particular machine
 * will usually take one microsecond to execute 10 iterations of this workload.
 */
static void update_average_slope(const int workload, const int verbose)
{
	unsigned long run_nsec, iterations_per_second;
	double estimated_slope;
	int i;
	double sum, average_slope;
	struct workload *w = get_workload_struct(workload);

	//initialize the workload global data
	if (w->init_global)
		w->init_global();

	//run workload once for a fairly large number of iterations, and use this
	//number of iterations to estimate a very loose slope number
	run_nsec = measure_workload(w, DEFAULT_SLOPE_ITERS);
	estimated_slope = (1000.0 * DEFAULT_SLOPE_ITERS) / run_nsec;	//one thousand is the conversion from nanoseconds to microseconds

	//refine this iteration once, and warm up the workload by running it for approximately a tenth of a second
	iterations_per_second = MILLION * estimated_slope;
	run_nsec = measure_workload(w, iterations_per_second);
	estimated_slope = (1000.0 * iterations_per_second) / run_nsec;	//one thousand is the conversion from nanoseconds to microseconds

	iterations_per_second = MILLION * estimated_slope;
	sum = 0;
	for (i = 0; i < SLOPE_GEN_SAMPLES; i++) {
		double slope;
		unsigned long iterations =
		    (SLOPE_GEN_MIN_TIME +
		     i * ((SLOPE_GEN_MAX_TIME - SLOPE_GEN_MIN_TIME) /
			  SLOPE_GEN_SAMPLES)) * estimated_slope;
		run_nsec = measure_workload(w, iterations);
		slope = (1000.0 * iterations) / run_nsec;
		sum += slope;
		estimated_slope = sum / (i + 1);
		if (verbose) {
			printf
			    ("\t\titers: %ld, us: %ld, slope: %f (average: %f)\n",
			     iterations, run_nsec / 1000, slope,
			     estimated_slope);
		}
	}

	//tear down the global workload data
	if (w->cleanup_global)
		w->cleanup_global();

	average_slope = sum / SLOPE_GEN_SAMPLES;

	if (put_workload_slope(workload, TIMING_AVERAGE, average_slope))
		fatal_error("Unable to write average slope");
}

/*
 * Update the slopes for this workload. If force is zero (false), only
 * regenerate the slope value if it is not already defined. If force is true,
 * regenerate all the slopes for this workload.
 */
void update_slope(int workload, const int force, const int verbose)
{
	const char *workload_name = get_workload_name(workload);
	double average_slope, wcet_slope;

	//if the slope is not accessible, complain about access errors
	if (!workload_slope_file_accessible(workload, TIMING_AVERAGE) ||
	    !workload_slope_file_accessible(workload, TIMING_WCET)) {
		fatal_error
		    ("Could not access the slope files. Check to make sure you are running as root (or sudo).");
	}
	//get the existing slopes if they do, in fact, exist
	average_slope = get_workload_slope(workload, TIMING_AVERAGE);
	wcet_slope = get_workload_slope(workload, TIMING_WCET);

	printf("Updating the slope for the %s workload:\n", workload_name);

	if (average_slope <= 0.0 || force) {
		printf("\tRecalculating average slope...\n");
		update_average_slope(workload, verbose);
		average_slope = get_workload_slope(workload, TIMING_AVERAGE);
		printf("\tAverage slope calculated to be: %f\n", average_slope);
	} else {
		printf
		    ("\tAverage slope for the %s workload has already been calculated to be %f. Skipping.\n",
		     workload_name, average_slope);
	}

	if (wcet_slope <= 0.0 || force) {
		printf("\tRecalculating WCET slope...\n");
		update_wcet_slope(workload, verbose);
		wcet_slope = get_workload_slope(workload, TIMING_WCET);
		printf("\tWCET slope calculated to be: %f\n", wcet_slope);
	} else {
		printf
		    ("\tWCET slope for the %s workload has already been calculated to be %f. Skipping.\n",
		     workload_name, wcet_slope);
	}
}
