/***************************************************************************
 *   Copyright (C) 2009-2011 Virginia Tech Real-Time Systems Lab           *
 *                                                                         *
 *   Original version written by Matthew Dellinger                         *
 *   mdelling@vt.edu                                                       *
 *                                                                         *
 *   Re-written by Aaron Lindsay                                           *
 *   aaron@aclindsay.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "task.h"

/*
 * Given a comma-separated list of processor IDs in string cpus[], initalize
 * a scheduling domain mask which contains all these processors, and if it is not
 * already in the global list of scheduling domains, add it.
 */
static void initialize_task_cpus(struct test *tester, struct task *t,
				 char cpus[])
{
	char tmp[32];
	char *c;
	int i, j = 0;
	int cpu;
	unsigned long mask = 0;

	//zero the task's cpu affinity mask
	MASK_ZERO(t->cpu_mask);

	/*
	 * If we specified 'all' of the cpus, add them all and return.
	 */
	if (cpus[0] == 'a' && cpus[1] == 'l' && cpus[2] == 'l') {
		for (i = 0; i < tester->num_processors; i++) {
			MASK_SET(mask, i);	//add cpu to domain mask
			MASK_SET(t->cpu_mask, i);	//add cpu to task affinity mask
		}
	} else {
		/*
		 * If we specified some subset of the cpus, parse which ones and add them to
		 * the appropriate domain mask
		 */
		c = cpus;
		while (1) {
			i = 0;
			//read the next integer into tmp
			do {
				tmp[i++] = c[j++];
			} while (tmp[i - 1] >= '0' && tmp[i - 1] <= '9');
			tmp[i] = '\0';

			//read the cpu as an integer, and add it to the mask
			cpu = atoi(tmp);
			if (cpu >= tester->num_processors) {
				printf
				    ("Error: The processor specified (%d) is larger than the number of procssors on this system(%d). Exiting.\n",
				     cpu, tester->num_processors);
				exit(1);
			}
			MASK_SET(mask, cpu);	//add cpu to domain mask
			MASK_SET(t->cpu_mask, cpu);	//add cpu to task affinity mask

			//if the bad character wasn't a comma, we're done here
			if (c[j - 1] != ',')
				break;
		}
	}

	if (!mask)
		fatal_error("Empty domain mask");

	//add domain mask to list of masks if doesn't already exist
	//FIXME This doesn't enforce that scheduling domains are non-overlapping. Is this a problem?
	for (i = 0; i < tester->num_processors; i++) {
		if (tester->domain_masks[i] == mask)
			return;
		else if (tester->domain_masks[i] == 0) {
			tester->domain_masks[i] = mask;
			return;
		}
	}

	//if we reached here, it means that we ran off the end of the array of domain masks. Oops.
	fatal_error
	    ("You have more domain masks than processors. This isn't necessarily a bad thing, we just don't currently support it. Contact one of our developers (http://chronoslinux.org) and let us know you need this capability. Exiting.");
}

/*
 * Given a comma-separated list in string locks[], initalize this task's locks
 * array to point to those locks in the global locks list to which they refer.
 */
static void initialize_task_locks(struct test *tester, struct task *t,
				  char locks[])
{
	char tmp[32];
	char *l;
	int i, j = 0;
	int lock_no;

	//initialize our locks array
	t->my_locks =
	    (chronos_mutex_t **) malloc(sizeof(chronos_mutex_t *) *
					tester->num_locks);
	if (!t->my_locks)
		fatal_error("Failed to allocate memory.");

	/*
	 * If we specified 'all' of the locks, add them all and return.
	 */
	if (locks[0] == 'a' && locks[1] == 'l' && locks[2] == 'l') {
		t->num_my_locks = tester->num_locks;
		for (i = 0; i < tester->num_locks; i++)
			t->my_locks[i] = &tester->locks[i];
		return;
	}

	/*
	 * If we specified some subset of the locks, parse which ones and add them to
	 * the task's array of pointers to locks.
	 */
	l = locks;
	while (1) {
		i = 0;
		//read the next integer into tmp
		do {
			tmp[i++] = l[j++];
		} while (tmp[i - 1] >= '0' && tmp[i - 1] <= '9');
		tmp[i] = '\0';

		//read the lock as an integer, and get the pointer to the global locks array
		lock_no = atoi(tmp);
		t->my_locks[t->num_my_locks] = &tester->locks[lock_no];
		t->num_my_locks++;

		//if the bad character wasn't a comma, we're done here
		if (l[j - 1] != ',')
			break;
	}
}

static void set_period_ts(struct task *t)
{
	t->period_ts.tv_sec = t->period / MILLION;
	t->period_ts.tv_nsec =
	    t->period * THOUSAND - t->period_ts.tv_sec * MILLION * THOUSAND;
}

/*
 * Read the line for one task from the taskset file and create a task struct from it
 */
void init_task(struct test *tester, FILE * f)
{
	char cpus[1024], locks[1024];
int tmp;
	int num_fields, locks_present, hua_present;

	//initialize the task's struct
	struct task *t = new_task(tester);

	//get the first four fields
	num_fields =
	    fscanf(f, "%s %ld %ld %d ", cpus, &t->period, &t->exec_time,
		   &t->utility);
	if (num_fields != 4)
		fatal_error("Ill-formed taskset file. num_fields");

	set_period_ts(t);	//initialize the period as a timespec, since this is required to be passed into the syscall later

	initialize_task_cpus(tester, t, cpus);

	//attempt to read last two fields, one at a time
	//first, list of locks
	tmp = getc(f);
	if (tmp == EOF)
		goto add_task;
	else if (tmp == 'T' || tmp == '#') {
		ungetc(tmp, f);
		goto add_task;
	} else {
		ungetc(tmp, f);
		locks_present = fscanf(f, " %s ", locks);
		if (locks_present != 1)
			fatal_error("Ill-formed taskset file. locks_present");
		initialize_task_locks(tester, t, locks);
	}

	//if we got here, attempt to get HUA utility as well
	tmp = getc(f);
	if (tmp == EOF)
		goto add_task;
	else if (tmp < '0' || tmp > '9') {
		ungetc(tmp, f);
		goto add_task;
	} else {
		ungetc(tmp, f);
		hua_present = fscanf(f, " %d ", &t->hua_utility);
		if (hua_present != 1)
			fatal_error("Ill-formed taskset file. ungetc");
	}

 add_task:
	//add task to the beginning of the list
	t->next = tester->task_list;
	tester->task_list = t;
	tester->num_tasks++;
}
