/***************************************************************************
 *   Copyright (C) 2009-2011 Virginia Tech Real-Time Systems Lab           *
 *                                                                         *
 *   Written by Aaron Lindsay                                              *
 *   aaron@aclindsay.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef WORKLOAD_TYPE_H
#define WORKLOAD_TYPE_H

/*
 * This struct is what is used to hold the functions for a workload in a way so
 * they can be generically called without having to do a select-case every time
 * on the workload type.
 */
struct workload {
	/*
	 * The workload's name. This is used to calculate the slope filenames for this
	 * workload, in addition to serving as an identifier (may be printed in console
	 * messages about this workload).
	 */
	const char *name;

	/*
	 * Initialize the global state needed for this workload. This should include
	 * initializing anything that is not per-task. It is assumed that any such state
	 * will be stored in static variables in that workload's execution unit, so no
	 * pointer is returned as is the case for init_task, below.
	 */
	void (*init_global) ();

	/*
	 * Initialize the per-task state needed for this workload. Return a void pointer
	 * to any per-task data which needs to be passed in whenever do_work is called
	 * for a task.
	 */
	void *(*init_task) ();

	/*
	 * Do the reverse of init_global. Free any memory malloc-ed there, etc.
	 */
	void (*cleanup_global) ();

	/*
	 * Do the reverse of init_global. Free any memory malloc-ed there, etc. The
	 * pointer passed in is the same pointer returned by init_task earlier.
	 */
	void (*cleanup_task) (void *);

	/*
	 * This function is the "heart" of the workload. The pointer passed in is the
	 * same pointer returned earlier by init_task. The unsigned long argument is
	 * the number of iterations this workload should complete of whatever work it
	 * is doing. The word 'iterations' is used somewhat loosely - the main thing
	 * that matters is that the relationship between the number passed in and the
	 * amount of work done when calling this function must be linear. It is
	 * preferable that each 'iteration' take a fairly small amount of time (i.e.
	 * less than a microsecond, preferably 10ths of a microsecond, or even
	 * nanoseconds).
	 */
	void (*do_work) (void *, unsigned long);
};

#endif				/* WORKLOAD_TYPE_H */
