/***************************************************************************
 *   Copyright (C) 2009-2011 Virginia Tech Real-Time Systems Lab           *
 *                                                                         *
 *   Written by Aaron Lindsay                                              *
 *   aaron@aclindsay.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <time.h>

#include "../workload_type.h"

#ifndef LINKED_LIST_H
#define LINKED_LIST_H

#define LIST_SIZE 64		//kilobytes

struct node {
	int value;
	struct node *next;
};

struct linked_list_data {
	struct node *head;
	struct node *next_node;
	long num_elements;
};

/*
 * Insert node n into the list begininning at data->head in order of increasing
 * value.
 */
static void insert_node(struct linked_list_data *data, struct node *n)
{
	struct node *cur = data->head;
	struct node *last = 0;

	if (!data->head) {
		data->head = n;
		return;
	}

	while (cur && cur->value > n->value) {
		last = cur;
		cur = cur->next;
	}

	if (!cur) {		//means we hit the end of the list
		last->next = n;
	} else if (!last) {	//i.e. our value is less than head's
		n->next = data->head;
		data->head = n;
	} else {		//our value is > last's but <= cur's
		n->next = cur;
		last->next = n;
	}
}

/*
 * Initialize the linked list by creating enough elements to fill LIST_SIZE with
 * random values, then insert them in increasing order of value into the list.
 */
static void *linked_list_init_task()
{
	int i;
	struct linked_list_data *data = malloc(sizeof(struct linked_list_data));
	assert(data);

	srand(time(NULL));	//seed random number generator

	data->num_elements = (LIST_SIZE * 1024) / sizeof(struct node);
	data->head = 0;

	//insert all nodes
	for (i = 0; i < data->num_elements; i++) {
		struct node *n = (struct node *)malloc(sizeof(struct node));
		assert(n);
		n->value = (int)rand();
		n->next = 0;
		insert_node(data, n);
	}

	data->next_node = data->head;

	return data;
}

/*
 * Free the memory malloc-ed for the list in linked_list_init_task()
 */
static void linked_list_cleanup_task(void *ptr)
{
	struct linked_list_data *data = (struct linked_list_data *)ptr;

	while (data->head != 0) {
		struct node *tmp = data->head;
		data->head = data->head->next;
		free(tmp);
	}

	free(data);
}

/*
 * Walk the linked list, one element at a time, incrementing the list element's
 * value as we go. This makes sure we'll get both a memory read and write
 * operation for each iteration.
 */
static void linked_list_do_work(void *ptr, unsigned long iterations)
{
	unsigned long i;
	struct linked_list_data *data = (struct linked_list_data *)ptr;

	for (i = 0; i < iterations; i++) {
		data->next_node = data->next_node->next;
		if (!data->next_node)
			data->next_node = data->head;
		data->next_node->value++;
	}
}

static const char linked_list_name[] = "linked_list";
static struct workload linked_list_workload = {
	.name = linked_list_name,
	.init_global = 0,
	.init_task = linked_list_init_task,
	.cleanup_global = 0,
	.cleanup_task = linked_list_cleanup_task,
	.do_work = linked_list_do_work
};

#endif
